#include "tokens_type.h"

enum TokenClass {
    KEYWORD_NAME = 4,
    TOKENIZABLE = 8
};


struct TokenInfo {
    std::string text;
    unsigned short attributes;
};


struct TokenTables {
    std::unordered_map<TokenType, TokenInfo> allTokensInfo;
    std::unordered_map<std::string, TokenType> keywords;

    TokenType tokenize(std::string value, TokenType initial);

    bool is_keyword(TokenType token);

    void add(TokenType tokenType, const std::string &txt,
             unsigned short attribs = 0);

    TokenTables();
};


static TokenTables tokenTables;

void TokenTables::add(TokenType tokenType, const std::string &txt,
                      unsigned short attribs) {
    allTokensInfo[tokenType] = TokenInfo{txt, attribs};
    if (attribs & KEYWORD_NAME)
        keywords[txt] = tokenType;
}

TokenTables::TokenTables() {

    add(TokenType::IDENTIFIER, "identifier");
    add(TokenType::KEYWORD, "keyword", KEYWORD_NAME);

    // Literals (bool is not here, it has two keywords: false and true)
    add(TokenType::NUMBER, "number");
    add(TokenType::STRING, "string");
    add(TokenType::COLOR, "color");

    // Parentheses of all kinds
    add(TokenType::LPAREN, "(");
    add(TokenType::RPAREN, ")");
    add(TokenType::LBRACKET, "[");
    add(TokenType::RBRACKET, "]");
    add(TokenType::LBRACE, "{");
    add(TokenType::RBRACE, "}");
    add(TokenType::LT, "<");
    add(TokenType::GT, ">");

    // assignments
    add(TokenType::EQUALS, "=");

    // miscellaneous
    add(TokenType::DOT, ".");
    add(TokenType::COLON, ":");
    add(TokenType::SEMICOLON, ";");
    add(TokenType::COMMA, ",");
    add(TokenType::WHITESPACE, "whitespace");
    add(TokenType::COMMENT, "comment");

    add(TokenType::TRUE, "true", TOKENIZABLE);
    add(TokenType::FALSE, "false", TOKENIZABLE);

    add(TokenType::IF, "if", TOKENIZABLE);
    add(TokenType::ELSE, "else", TOKENIZABLE);

    add(TokenType::DEFAULT, "default", TOKENIZABLE);
    add(TokenType::NIL, "null", TOKENIZABLE);

    // all keywords
    add(TokenType::KW_NAME, "@name", TOKENIZABLE | KEYWORD_NAME);
    add(TokenType::KW_INFO, "@info", TOKENIZABLE | KEYWORD_NAME);
    add(TokenType::KW_HEADER, "@header", TOKENIZABLE | KEYWORD_NAME);
    add(TokenType::KW_COMPONENT_HEADER, "@componentHeader", TOKENIZABLE | KEYWORD_NAME);
    add(TokenType::KW_DIRECTORY, "@directory", TOKENIZABLE | KEYWORD_NAME);
    add(TokenType::KW_LIBRARY, "@library", TOKENIZABLE | KEYWORD_NAME);
    add(TokenType::KW_ADDRESS, "@address", TOKENIZABLE | KEYWORD_NAME);
    add(TokenType::KW_COMPONENT, "@component", TOKENIZABLE | KEYWORD_NAME);
    add(TokenType::KW_MESSAGES, "@messages", TOKENIZABLE | KEYWORD_NAME);
    add(TokenType::KW_INSTANCE_NAME, "@instanceName", TOKENIZABLE | KEYWORD_NAME);
    add(TokenType::KW_COUNT, "@count", TOKENIZABLE | KEYWORD_NAME);
    add(TokenType::KW_DISPLAY, "@display", TOKENIZABLE | KEYWORD_NAME);
    add(TokenType::KW_PIN, "@pin", TOKENIZABLE | KEYWORD_NAME);
    add(TokenType::KW_TOOLTIP, "@tooltip", TOKENIZABLE | KEYWORD_NAME);
    add(TokenType::KW_CONNECTION, "@connection", TOKENIZABLE | KEYWORD_NAME);
    add(TokenType::KW_ATTRIBUTE, "@attribute", TOKENIZABLE | KEYWORD_NAME);
    add(TokenType::KW_SOURCE, "@source", TOKENIZABLE | KEYWORD_NAME);
    add(TokenType::KW_POPUP, "@popup", TOKENIZABLE | KEYWORD_NAME);
    add(TokenType::KW_RULE, "@rule", TOKENIZABLE | KEYWORD_NAME);
    add(TokenType::KW_TITLE, "@title", TOKENIZABLE | KEYWORD_NAME);
    add(TokenType::KW_TEXT, "@text", TOKENIZABLE | KEYWORD_NAME);
    add(TokenType::KW_BUS, "@bus", TOKENIZABLE | KEYWORD_NAME);
    add(TokenType::KW_WIRES, "@wires", TOKENIZABLE | KEYWORD_NAME);
    add(TokenType::KW_ENUMERATED, "@enumerated", TOKENIZABLE | KEYWORD_NAME);
    add(TokenType::KW_WIRE, "@wire", TOKENIZABLE | KEYWORD_NAME);
    add(TokenType::KW_INSTANCE, "@instance", TOKENIZABLE | KEYWORD_NAME);
    add(TokenType::KW_SCHEMA, "@schema", TOKENIZABLE | KEYWORD_NAME);
    add(TokenType::KW_POSITION, "@position", TOKENIZABLE | KEYWORD_NAME);
    add(TokenType::KW_SIZE, "@size", TOKENIZABLE | KEYWORD_NAME);
    add(TokenType::KW_DISPLAY_NAME, "@displayName", TOKENIZABLE | KEYWORD_NAME);

    // All types
    add(TokenType::INT_TYPE, "int", TOKENIZABLE);
    add(TokenType::STRING_TYPE, "string", TOKENIZABLE);
    add(TokenType::BOOL_TYPE, "bool", TOKENIZABLE);
    add(TokenType::WIRE_TYPE, "wire", TOKENIZABLE);

    // WireNode types
    add(TokenType::WIRE_DEFAULT, "normal_wire", TOKENIZABLE);
    add(TokenType::WIRE_AND, "wired_and", TOKENIZABLE);
    add(TokenType::WIRE_OR, "wired_or", TOKENIZABLE);
    add(TokenType::R_WIRE, "r_wire", TOKENIZABLE);

    // ComponentNode types
    add(TokenType::CT_MEMORY, "memory", TOKENIZABLE);
    add(TokenType::CT_PROCESSOR, "processor", TOKENIZABLE);

    // PinNode type
    add(TokenType::PIN_IN, "in", TOKENIZABLE);
    add(TokenType::PIN_OUT, "out", TOKENIZABLE);
    add(TokenType::PIN_IN_OUT, "inOut", TOKENIZABLE);

    // ActionNode types
    add(TokenType::ERROR, "error", TOKENIZABLE);
    add(TokenType::WARNING, "warning", TOKENIZABLE);


    // WireNode settings types
    add(TokenType::HIDDEN, "hidden", TOKENIZABLE);
    add(TokenType::TERMINATE_WITH, "terminate_with", TOKENIZABLE);

    // Miscelaneous
    add(TokenType::END_OF_FILE, "eof");
}


const std::string &tokenTypeToString(TokenType tokenType) {
    return tokenTables.allTokensInfo[tokenType].text;
}

TokenType TokenTables::tokenize(std::string value, TokenType initial) {
    for (auto &[key, param]: allTokensInfo) {
        if (param.attributes & TOKENIZABLE && param.text == value) {
            return key;
        }
    }
    return initial;
}

bool TokenTables::is_keyword(TokenType token) {
    for (auto &[key, param]: allTokensInfo) {
        if (param.attributes & KEYWORD_NAME && key == token) {
            return true;
        }
    }
    return false;
}

TokenType from_token(std::string value, TokenType initial) {
    return tokenTables.tokenize(value, initial);
}

bool is_keyword(TokenType tokenType) {
    return tokenTables.is_keyword(tokenType);
}